/*
 * The PyHST program is Copyright (C) 2002-2011 of the
 * European Synchrotron Radiation Facility (ESRF) and
 * Karlsruhe Institute of Technology (KIT).
 *
 * PyHST is free software: you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the
 * Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * hst is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#define BLOCK_SIZE 16

#if SLICE_BLOCK == 4
# define vfloat float4
# define tex2d(a,b,c) read_imagef(a, b, c)
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; slice[2 * slice_size + idy*slice_width + idx] = sum.z; slice[3 * slice_size + idy*slice_width + idx] = sum.w; })
#elif SLICE_BLOCK == 2
# define vfloat float2
# define tex2d(a,b,c) read_imagef(a, b, c).xy
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; })
#else
# define vfloat float
# define tex2d(a,b,c) read_imagef(a, b, c).x
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum; })
#endif


const sampler_t volumeSampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP | CLK_FILTER_LINEAR; 
const sampler_t nearestSampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP | CLK_FILTER_NEAREST; 


__kernel void hst_filter(__global float *data, __global float *filter)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int id = idy*get_global_size(0) + idx;
    data[id] = data[id] * filter[idx];
}


__kernel void hst_normalize_idft(__global float *data)
{
    const int idx = get_global_id(0);
    const int dim_fft = get_global_size(0);
    data[2*idx] = data[2*idx] / dim_fft;
}

__kernel void hst_spread(__global float *out, const int dpitch, __global float *in, const int spitch, const int num_bins)
{
    float tmp1, tmp2;
    const int idx = get_global_id(0);
    const int s_idy = get_global_id(1);
    const int d_idy = get_group_id(1) * get_local_size(1) + get_local_id(1);

    int pos = 2 * s_idy * spitch + idx;

    tmp1 = (idx < num_bins)?in[pos]:0;
    tmp2 = (idx < num_bins)?in[pos + spitch]:0;

    pos = 2 * (d_idy * dpitch + idx);

    out[pos] = tmp1;
    out[pos + 1] = tmp2;
}

__kernel void hst_pack(__global float *in, const int spitch,  __global vfloat *out, const int dpitch, const int num_proj)
{
    float tmp1, tmp2;
    const int idx = get_global_id(0);
    const int d_idy = get_global_id(1);
    const int s_idy = get_group_id(1) * get_local_size(1) + get_local_id(1);
    const int proj = 2 * d_idy;

    int pos = 2 * (s_idy * spitch + idx);

    tmp1 = in[pos];
    tmp2 = in[pos + 1];

#if SLICE_BLOCK == 1
    pos = proj * dpitch + idx ;

    out[pos] = tmp1;
    out[pos + dpitch] = tmp2;
#else
    if (proj < num_proj) {
	pos = proj * dpitch + idx ;

	out[pos].x = tmp1;
	out[pos + dpitch].x = tmp2;
    } 
    else if (proj < 2 * num_proj) {
	pos = (proj - num_proj) * dpitch + idx;
	out[pos].y = tmp1;
	out[pos + dpitch].y = tmp2;
    }
#endif
#if SLICE_BLOCK == 4
    else if (proj < 3 * num_proj) {
	pos = (proj - 2 * num_proj) * dpitch + idx;
	out[pos].z = tmp1;
	out[pos + dpitch].z = tmp2;
    }
    else if (proj < 4 * num_proj) {
	pos = (proj - 3 * num_proj) * dpitch + idx;
	out[pos].w = tmp1;
	out[pos + dpitch].w = tmp2;
    }
#endif
}



__kernel void hst_backproject_gpu_nearest(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all,
    __local float *shared) {

    __local float *cache = shared;
    __local float *cache_subh = shared + 48 * BLOCK_SIZE;			// cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1]
    __local float *cache_sin = cache_subh +  BLOCK_SIZE * (2*BLOCK_SIZE + 1);	// cache_sin[2 * BLOCK_SIZE_Y]
/*
    __local float cache[2 * BLOCK_SIZE * 24];
    __local float cache_subh[BLOCK_SIZE * (2 * BLOCK_SIZE + 1)];
    __local float cache_sin[2 * BLOCK_SIZE];
*/
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
//    const int bidy = get_group_id(1) * BLOCK_SIZE;
    const int bidy = get_global_id(1) - get_local_id(1);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    float4 all1, all2;

    int minh1, minh2;

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE + (tidx & 7);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)floor(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)floor(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

        cache[tidy2 * 24 + tidx] = read_imagef(sinogram, volumeSampler, (float2)(minh1 + tidx, proj + 0.5f)).x;
        cache[tidy21 * 24 + tidx] = read_imagef(sinogram, volumeSampler, (float2)(minh2 + tidx, proj + 1.5f)).x;
        cache[tidy2p * 24 + pos] = read_imagef(sinogram, volumeSampler, (float2)((prj?minh2:minh1) + pos, proj + prj + 0.5f)).x;

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int j = 2 * i;
	    minh1 = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    res += cache[j * 24 + minh1];

	    ++j;
	    minh1 = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    res += cache[j * 24 + minh1];
	}

        barrier(CLK_LOCAL_MEM_FENCE);

    }

    slice[idy*slice_width + idx] = res;
}

__kernel void hst_backproject_gpu_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][6 * BLOCK_SIZE_X];
    __local float *cache_subh = cache + 6 * BLOCK_SIZE * BLOCK_SIZE;		// cache_subh[BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __local float *cache_sin = cache_subh + BLOCK_SIZE * (BLOCK_SIZE + 1);	// cache_sin[BLOCK_SIZE_Y]

    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
//    const int bidy = get_group_id(1) * BLOCK_SIZE;
    const int bidy = get_global_id(1) - get_local_id(1);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    
//#pragma unroll 8
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	cache_subh[tidx * (BLOCK_SIZE + 1) + tidy] = 4 * (all.z + x * all.x - minh);
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;

#pragma unroll 6
	for (int i = 0; i < 6; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * 6 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + ((float)pos)/4, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    minh = cache_subh[tidx * (BLOCK_SIZE + 1) + i] - y * cache_sin[i];
	    res += cache[i * 6 * BLOCK_SIZE + minh];
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[idy*slice_width + idx] = res;
}

__kernel void hst_backproject_gpu_linear(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float2 *cache = (__local float2*)shared;						// cache[2 * BLOCK_SIZE_Y][24];
    __local float *cache_subh = shared + 96 * BLOCK_SIZE;					// cache_subh[BLOCK_SIZE_X][2 * BLOCK_SIZE_Y + 1]
    __local float *cache_sin = cache_subh +  BLOCK_SIZE * (2*BLOCK_SIZE + 1);			// cache_sin[2 * BLOCK_SIZE_Y]


    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * BLOCK_SIZE;
//    const int bidy = get_group_id(1) * BLOCK_SIZE;
    const int bidy = get_global_id(1) - get_local_id(1);

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = bx + tidx;
    const float y = by + tidy;
    
    const int slice_width = get_global_size(0);

    float res = 0;
    float4 all1, all2;

    float minh;
    int minh1, minh2;

    const int prj = tidx >> 3;
    const int pos = BLOCK_SIZE + (tidx & 7);

    const int tidy2 = tidy << 1;
    const int tidy21 = tidy2 + 1;
    const int tidy2p = tidy2 + prj;

    for (int proje=0; proje<num_proj; proje += 2 * BLOCK_SIZE) {
	int proj = proje + tidy2;
	
        all1 = c_all[proj];
        all2 = c_all[proj + 1];
	minh1 = (int)floor(all1.z + bx * all1.x - by * all1.y + all1.w);
	minh2 = (int)floor(all2.z + bx * all2.x - by * all2.y + all2.w);
		
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy2] = all1.z + x * all1.x - minh1;
	cache_subh[tidx * (2 * BLOCK_SIZE + 1) + tidy21] = all2.z + x * all2.x - minh2;
	
	if (tidx == 0) {
	    cache_sin[tidy2] = all1.y;
	    cache_sin[tidy21] = all2.y;
	}

        cache[tidy2 * 24 + tidx].x = read_imagef(sinogram, volumeSampler, (float2)(minh1 + tidx, proj + 0.5f)).x;
        cache[tidy21 * 24 + tidx].x = read_imagef(sinogram, volumeSampler, (float2)(minh2 + tidx, proj + 1.5f)).x;
        cache[tidy2p * 24 + pos].x = read_imagef(sinogram, volumeSampler, (float2)((prj?minh2:minh1) + pos, proj + prj + 0.5f)).x;

        barrier(CLK_LOCAL_MEM_FENCE);

    	cache[tidy2 * 24 + tidx].y = cache[tidy2 * 24 + tidx + 1].x - cache[tidy2 * 24 + tidx].x;
	cache[tidy21 * 24 + tidx].y = cache[tidy21 * 24 + tidx + 1].x - cache[tidy21 * 24 + tidx].x;
	if (pos < 23) {
	    cache[tidy2p * 24 + pos].y = cache[tidy2p * 24 + pos + 1].x - cache[tidy2p * 24 + pos].x;
	} else {
	    cache[tidy2p * 24 + pos].y = 0;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
	
#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int j = 2 * i;
	    float2 v;
	    
	    minh = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j * 24 + minh1];
	    res += v.x + (minh - minh1) * v.y;

	    ++j;
	    minh = cache_subh[tidx * (2 * BLOCK_SIZE + 1) + j] - y * cache_sin[j];
	    minh1 = (int)minh;
	    v = cache[j * 24 + minh1];
	    res += v.x + (minh - minh1) * v.y;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[idy*slice_width + idx] = res;
}


__kernel void hst_backproject_fermi_nearest(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X];
    __local float *cache_subh = cache + 3 * BLOCK_SIZE * BLOCK_SIZE;		// cache_subh[2 * BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __local float *cache_sin = cache_subh + 2 * BLOCK_SIZE * (BLOCK_SIZE + 1);	// cache_sin[BLOCK_SIZE_Y]

    const int idx = 2 * get_global_id(0);
    const int idy = 2 * get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = idx + off_x;
    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float2 res[2] = {{0.f, 0.f}, {0.f, 0.f}};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + tidy] = subh;
	cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * 3 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int4 minh = {
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i]
	    };

	    res[0].x += cache[i * 3 * BLOCK_SIZE + minh.x];
	    res[0].y += cache[i * 3 * BLOCK_SIZE + minh.y];
	    res[1].x += cache[i * 3 * BLOCK_SIZE + minh.z];
	    res[1].y += cache[i * 3 * BLOCK_SIZE + minh.w];
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res[0].x;
    slice[slice_width * (idy    ) + idx + 1] = res[0].y;
    slice[slice_width * (idy + 1) + idx] = res[1].x;
    slice[slice_width * (idy + 1) + idx + 1] = res[1].y;
}



__kernel void hst_backproject_fermi_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];
    __local float *cache_subh = cache + 12 * BLOCK_SIZE * BLOCK_SIZE;		// cache_subh[2 * BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __local float *cache_sin = cache_subh + 2 * BLOCK_SIZE * (BLOCK_SIZE + 1);	// cache_sin[BLOCK_SIZE_Y]

    const int idx = 2 * get_global_id(0);
    const int idy = 2 * get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = idx + off_x;
    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float2 res[2] = {{0.f, 0.f}, {0.f, 0.f}};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = 4 * (all.z + x * all.x - minh);
	cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + tidy] = subh;
	cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + tidy] = subh + 4 * all.x;
	if (tidx == 0) cache_sin[tidy] = 4 * all.y;

#pragma unroll 12
	for (int i = 0; i < 12; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * 12 * BLOCK_SIZE + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + ((float)pos)/4, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    int4 minh = {
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i]
	    };

	    res[0].x += cache[i * 12 * BLOCK_SIZE + minh.x];
	    res[0].y += cache[i * 12 * BLOCK_SIZE + minh.y];
	    res[1].x += cache[i * 12 * BLOCK_SIZE + minh.z];
	    res[1].y += cache[i * 12 * BLOCK_SIZE + minh.w];
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res[0].x;
    slice[slice_width * (idy    ) + idx + 1] = res[0].y;
    slice[slice_width * (idy + 1) + idx] = res[1].x;
    slice[slice_width * (idy + 1) + idx + 1] = res[1].y;
}


__kernel void hst_backproject_fermi_linear(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float2 *cache = (__local float2*)shared;						// cache[BLOCK_SIZE_Y][3 * BLOCK_SIZE_X + 1];
    __local float *cache_subh = shared + 6 * BLOCK_SIZE * BLOCK_SIZE + 2 * BLOCK_SIZE;		// cache_subh[2 * BLOCK_SIZE_X][BLOCK_SIZE_Y + 1];
    __local float *cache_sin = cache_subh + 2 * BLOCK_SIZE * (BLOCK_SIZE + 1);			// cache_sin[BLOCK_SIZE_Y]

    const int idx = 2 * get_global_id(0);
    const int idy = 2 * get_global_id(1);

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const float x = idx + off_x;
    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	int minh = (int)floor(all.z + bx * all.x - by * all.y + all.w);

	float subh = (all.z + x * all.x - minh);
	cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + tidy] = subh;
	cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + tidy] = subh + all.x;
	if (tidx == 0) cache_sin[tidy] = all.y;

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].y = cache[tidy * (3 * BLOCK_SIZE + 1) + pos + 1].x - cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x;
	}


        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
	    float2 c;
	    float4 v1, v2;
	    float4 minh = {
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y    ) * cache_sin[i],
		cache_subh[(2 * tidx    ) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i],
		cache_subh[(2 * tidx + 1) * (BLOCK_SIZE + 1) + i] - (y + 1) * cache_sin[i]
	    };

	    c = cache[i * (3 * BLOCK_SIZE + 1) + (int)minh.x];
	    v1.x = c.x; v2.x = c.y;
	    c = cache[i * (3 * BLOCK_SIZE + 1) + (int)minh.y];
	    v1.y = c.x; v2.y = c.y;
	    c = cache[i * (3 * BLOCK_SIZE + 1) + (int)minh.z];
	    v1.z = c.x; v2.z = c.y;
	    c = cache[i * (3 * BLOCK_SIZE + 1) + (int)minh.w];
	    v1.w = c.x; v2.w = c.y;
	    
	    res += v1 + (minh - floor(minh)) * v2;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res.x;
    slice[slice_width * (idy    ) + idx + 1] = res.y;
    slice[slice_width * (idy + 1) + idx] = res.z;
    slice[slice_width * (idy + 1) + idx + 1] = res.w;
}



/*
__kernel void hst_pack(__global float *in, __write_only image2d_t out, const int num_bins)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int dpitch = get_global_size(0)*2;

    if (idx < num_bins) {
//        out[idy*num_bins + idx] = in[idy*dpitch + idx*2];
	int2 pos = {idx, idy};
	float4 val = {in[idy*dpitch + idx*2], 0, 0, 0};
	write_imagef(out, pos, val);
    }
}
*/

__kernel void hst_backproject_gpu(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *all_table)
{
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);

    const int slice_width = get_global_size(0);	// Scale by PPT if necessary
    const int slice_size = get_global_size(1) * slice_width; // Scale by PPT if necessary

    float h;
    float bx = idx + off_x;
    float by = -(idy + off_y);
    vfloat sum = 0.0f;

    for(int proj = 0; proj < num_proj; proj++) {
	h = mad(by, all_table[proj].y, mad(bx, all_table[proj].x, all_table[proj].z)); 
//	sum += read_imagef(sinogram, volumeSampler, (float2)(h, proj + 0.5f)).x;
	sum += tex2d(sinogram, volumeSampler, (float2)(h, proj + 0.5f));
    }

    SET_SLICE(slice, idx, idy, sum);
}

//HST_OPENCL_CONST_HACK - do not edit, the string is used to remove following code if hack is disabled

__kernel void hst_backproject_gpu_const(const int num_proj,
    const int num_bins,
    const float off_x,
    const float off_y,
    __read_only image2d_t sinogram,
    __global float *slice)
{/*
    const int idx = get_global_id(0);
    const int idy = get_global_id(1);
    const int slice_width = get_global_size(0);

    float h;
    float bx = idx + off_x;
    float by = -(idy + off_y);
    float sum = 0.0f;

    for(int proj = 0; proj < num_proj; proj++) {
	h = mad(by, as_float(sin_itable[proj]), mad(bx, as_float(cos_itable[proj]), as_float(axis_itable[proj]))); 
	sum += read_imagef(sinogram, volumeSampler, (float2)(h, proj + 0.5f)).x;
    }
    slice[idy*slice_width + idx] = sum;*/
}


__kernel void hst_backproject_kepler_tex(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all
) {
    float h;
    vfloat res[4][4] = {{0}};

    volatile __local vfloat buf[16][16];
    __local vfloat fin[16][16];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * get_local_size(0);
    const int bidy = get_group_id(1) * get_local_size(1);

    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int proj_major = tidy / 4;
    const int xy_major = tidy % 4;
    const int y_major = xy_major / 2;
    const int x_major = xy_major % 2;

    const int proj = 4 * proj_major + proj_minor;
    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const float x = bidx + sidx + off_x;
    const float y = bidy + sidy + off_y;
    const float projf = proj + 0.5f;

    for (int proje=0; proje<num_proj; proje+=16) {
        const float4 all = c_all[proje+proj];
//	const float4 all = {700.f, -0.5f, 0.8f, 0.f};
        h = all.z + x * all.x - y * all.y;
//        h = mad(x, all.x, mad(y, -all.y, all.z));

#pragma unroll 4
        for (int i = 0; i < 4; i++) {
#pragma unroll 4
	    for (int j = 0; j < 4; j++) {
	        float subh = h + 4 * j * all.x - 4 * i * all.y;
		res[i][j] += tex2d(sinogram, volumeSampler, (float2)(subh, proje + projf));
	    }
	}
//    	barrier(CLK_LOCAL_MEM_FENCE);
    }

#pragma unroll 4
    for (int i = 0; i < 4; i++) {
#pragma unroll 4
	for (int j = 0; j < 4; j++) {
            buf[4 * sidy + sidx][proj] = res[i][j];

    	    barrier(CLK_LOCAL_MEM_FENCE);

            for (int k=8; k>=1; k/=2) {
        	buf[tidy][tidx] += buf[tidy][tidx+k];
	    }
            vfloat val = buf[tidy][0];

            const int rx = 4 * j + tidy%4;
            const int ry = 4 * i + tidy/4;

            if (!tidx) {
		fin[ry][rx] = val;
            }

    	    barrier(CLK_LOCAL_MEM_FENCE);
        }
    }

    const int idx = bidx + tidx;
    const int idy = bidy + tidy;
    const int slice_width = get_global_size(0);	// Scale by PPT if necessary
    const int slice_size = get_global_size(1) * slice_width; // Scale by PPT if necessary

    SET_SLICE(slice, idx, idy, fin[tidy][tidx]);
//    SET_SLICE(slice, idx, idy, ((float2){0,1}));
//    slice[get_global_size(0) * idy + idx] = fin[tidy][tidx];
}






/* Direct modes may have higher error due to double interpolation. We took not the numbers
on integer positions, but interpolated numbers as it is convinient and only have the integer 
distance in between. */

#undef subh_cache
#define subh_cache(i, subh) cache[i * (3 * BLOCK_SIZE + 1) + (int)(subh)]
__kernel void hst_backproject_direct_nearest(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;


    const int slice_width = 2 * get_global_size(0);

    float res[4] = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);


#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float4 all = - c_all[proje + i];
            float h[4];

            //float subh = (-sidx * all.x - sidy * all.y) - all.w;
            h[0] = mad(sidx, all.x, mad(sidy, all.y, all.w));
            h[1] = h[0] - all.x;
            h[2] = h[0] + all.y;
            h[3] = h[1] + all.y;
            
#pragma unroll 4
            for (int j = 0; j < 4; j++)
                res[j] += subh_cache(i, h[j]);
        }
        
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy    ) + idx] = res[0];
    slice[slice_width * (idy    ) + idx + 1] = res[1];
    slice[slice_width * (idy + 1) + idx    ] = res[2];
    slice[slice_width * (idy + 1) + idx + 1] = res[3];
}

#undef subh_cache
#define subh_cache(i, subh) cache[i * (12 * BLOCK_SIZE + 1) + (int)(subh)]
//#define subh_cache(i, subh) (i * 12 * BLOCK_SIZE + (int)(subh))
__kernel void hst_backproject_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int tidx2 = 2 * get_local_id(0);
    const int tidy2 = 2 * get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));


        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = sbidx * 8 + 2 * stidx;
    const int sidy = sbidy * 8 + 2 * stidy;
    
    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    
    const int slice_width = 2 * get_global_size(0);

    float2 res[2] = {{0.f, 0.f}, {0.f, 0.f}};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	for (int i = 0; i < 12; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (12 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float4 all = 4 * c_all[proje + i];

            float subh = (sidx * all.x - sidy * all.y) - all.w;
            float subhx = subh + all.x;

	    res[0].x += subh_cache(i, subh);
	    res[0].y += subh_cache(i, subhx);
	    res[1].x += subh_cache(i, subh - all.y);
	    res[1].y += subh_cache(i, subhx - all.y);
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }
    
    slice[slice_width * (idy    ) + idx] = res[0].x;
    slice[slice_width * (idy    ) + idx + 1] = res[0].y;
    slice[slice_width * (idy + 1) + idx] = res[1].x;
    slice[slice_width * (idy + 1) + idx + 1] = res[1].y;
}

#undef subh_cache
#define subh_cache(i, subh) cache[i * (3 * BLOCK_SIZE + 1) + (int)(subh)]
__kernel void hst_backproject_direct_linear(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
#ifdef HST_TEXTURE_READS
    __read_only image2d_t sinogram,
#else
    __global __read_only float *sinogram,
#endif
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {
    __local float2 *cache = (__local float2 *)shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 4 + 1 * stidx);
    const int sidy = sbidy * 4 + 1 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;


    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
#ifdef HST_TEXTURE_READS
	float minh = all.z + bx * all.x - by * all.y + all.w + 0.5f;
#else
	int minh = all.z + bx * all.x - by * all.y + all.w;
#endif

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
#ifdef HST_TEXTURE_READS
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = read_imagef(sinogram, nearestSampler, (float2)(minh + pos, proj + 0.5f)).x;
#else
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = sinogram[proj * num_bins + minh + pos];
#endif

	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].y = cache[tidy * (3 * BLOCK_SIZE + 1) + pos + 1].x - cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float2 v;
            float4 v1, v2;
            float4 all = - c_all[proje + i];
            float4 h;

            //float subh = (-sidx * all.x - sidy * all.y) - all.w;
            h.x = mad(sidx, all.x, mad(sidy, all.y, all.w));
            h.y = h.x - BLOCK_SIZE * all.x;
            h.z = h.x + BLOCK_SIZE * all.y;
            h.w = h.y + BLOCK_SIZE * all.y;
            
            float4 ih = floor(h);
            float4 diff = h - ih;

            v = subh_cache(i, ih.x); v1.x = v.x; v2.x = v.y;
            v = subh_cache(i, ih.y); v1.y = v.x; v2.y = v.y;
            v = subh_cache(i, ih.z); v1.z = v.x; v2.z = v.y;
            v = subh_cache(i, ih.w); v1.w = v.x; v2.w = v.y;

            res += v1  + diff * v2;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy             ) + idx             ] = res.x;
    slice[slice_width * (idy             ) + idx + BLOCK_SIZE] = res.y;
    slice[slice_width * (idy + BLOCK_SIZE) + idx             ] = res.z;
    slice[slice_width * (idy + BLOCK_SIZE) + idx + BLOCK_SIZE] = res.w;
}

#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_amd_direct_nearest(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float *cache = shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = - (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx - sidx;
    const int idy = bidy + sidy;

    const int slice_width = 2 * get_global_size(0);

    float res[4] = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
	float minh = all.z + bx * all.x - by * all.y + all.w;

#pragma unroll 12
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + pos, proj + 0.5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);


#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float4 all = - c_all[proje + i];
            float h[4];

            //float subh = (-sidx * all.x - sidy * all.y) - all.w;
            h[0] = mad(49.f, i, mad(sidx, all.x, mad(sidy, all.y, all.w)));
            h[1] = h[0] - all.x;
            h[2] = h[0] + all.y;
            h[3] = h[1] + all.y;
            
#pragma unroll 4
            for (int j = 0; j < 4; j++)
                res[j] += subh_cache(h[j]);
        }
        
        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy    ) + idx] = res[0];
    slice[slice_width * (idy    ) + idx + 1] = res[1];
    slice[slice_width * (idy + 1) + idx    ] = res[2];
    slice[slice_width * (idy + 1) + idx + 1] = res[3];
}


#undef subh_cache
#define subh_cache(subh) cache[subh]
__kernel void hst_backproject_amd_direct_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {
    __local float *mcache = shared;
    __local float *cache = shared + BLOCK_SIZE * BLOCK_SIZE;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 8;
    const int stidy = tidx / 8;
    
    const int sidx = (sbidx * 8 + stidx);
    const int sidy = sbidy * 8 + 4 * stidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + off_x;
    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    const int num_blocks = (num_proj / BLOCK_SIZE) +  ((num_proj % BLOCK_SIZE)?1:0);
    
    for (int proj_block = 0; proj_block < num_blocks; proj_block += BLOCK_SIZE) {
        const int proj_offset = proj_block * BLOCK_SIZE;
        float4 all = c_all[proj_offset + BLOCK_SIZE * tidy + tidx];

	float minh = floor(all.z + bx * all.x - by * all.y + all.w);
	mcache[BLOCK_SIZE * tidy + tidx] = 193.f * tidx - 4 * minh;

        barrier(CLK_LOCAL_MEM_FENCE);

        const int sub_blocks = min(BLOCK_SIZE, num_blocks - proj_block);
        
        for (int p = 0; p < sub_blocks; p++) {
            const int proje = proj_offset + p * BLOCK_SIZE;
	    const int proj = proje + tidy;
	    
	    float minh = 193.f * tidy - mcache[p * BLOCK_SIZE + tidy];

#pragma unroll 12
	    for (int i = 0; i < 12; i++) {
	        int pos = i * BLOCK_SIZE + tidx;
	        cache[tidy * (12 * BLOCK_SIZE + 1) + pos] = read_imagef(sinogram, volumeSampler, (float2)(0.25f * (minh + pos), proj + 0.5f)).x;
	    }

            barrier(CLK_LOCAL_MEM_FENCE);
            
#pragma unroll 16
	    for (int i = 0; i < BLOCK_SIZE; i++) {
//            float4 all = -4 * c_all[proje + i];
//            float subh = mad(193.f, i, mad(sidx, all.x, mad(sidy, all.y, all.w)));
              float4 all = c_all[proje + i]; all.y = -all.y;
//	        float minh = 193 * i - 4 * floor(all.z + bx * all.x + by * all.y + all.w);

	        float minh = mcache[p * BLOCK_SIZE + i];

                float subh = mad(4,  mad(x, all.x, mad(y, all.y, all.z)),  minh);
	        res.x += subh_cache((int)subh);
	        res.y += subh_cache((int)mad(4.f, all.y, subh));
	        res.z += subh_cache((int)mad(8.f, all.y, subh));
	        res.w += subh_cache((int)mad(12.f, all.y, subh));
	    }

            barrier(CLK_LOCAL_MEM_FENCE);
        }
    }

    slice[slice_width * (idy    ) + idx] = res.x;
    slice[slice_width * (idy + 1) + idx] = res.y;
    slice[slice_width * (idy + 2) + idx] = res.z;
    slice[slice_width * (idy + 3) + idx] = res.w;
}

#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]
__kernel void hst_backproject_amd_direct_linear(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
#ifdef HST_TEXTURE_READS
    __read_only image2d_t sinogram,
#else
    __global __read_only float *sinogram,
#endif
    __global float *slice,
    __constant float4 *c_all, 
    __local float *shared
    ) {

    __local float2 *cache = (__local float2 *)shared;						// cache[BLOCK_SIZE_Y][12 * BLOCK_SIZE_X + 1];

    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = 2 * get_group_id(0) * BLOCK_SIZE;
//    const int bidy = 2 * get_group_id(1) * BLOCK_SIZE;
    const int bidy = 2 * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

        // This is subblocks within block (big block is 16x16, subblock is 4x4 completely contained in a single wrap)
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;
    
    const int sidx = (sbidx * 8 + 2 * stidx);
    const int sidy = sbidy * 8 + 2 * stidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + off_x;
    const float y = idy + off_y;
    
    const int slice_width = 2 * get_global_size(0);

    float4 res = {0.f, 0.f, 0.f, 0.f};
    
    for (int proje=0; proje<num_proj; proje += BLOCK_SIZE) {
	int proj = proje + tidy;
        float4 all = c_all[proj];
#ifdef HST_TEXTURE_READS
	float minh = floor(all.z + bx * all.x - by * all.y + all.w) + 0.5f;
#else
	int minh = floor(all.z + bx * all.x - by * all.y + all.w);
#endif


#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
#ifdef HST_TEXTURE_READS
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = read_imagef(sinogram, nearestSampler, (float2)(minh + pos, proj + 0.5f)).x;
#else
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x = sinogram[proj * num_bins + minh + pos];
#endif
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 3
	for (int i = 0; i < 3; i++) {
	    int pos = i * BLOCK_SIZE + tidx;
	    cache[tidy * (3 * BLOCK_SIZE + 1) + pos].y = cache[tidy * (3 * BLOCK_SIZE + 1) + pos + 1].x - cache[tidy * (3 * BLOCK_SIZE + 1) + pos].x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);

#pragma unroll 16
	for (int i = 0; i < BLOCK_SIZE; i++) {
            float2 v;
            float4 v1, v2;
            float4 all = c_all[proje + i]; all.y = - all.y;
            float4 h;
//	    float minh = all.z + bx * all.x - by * all.y + all.w;
	    float minh = floor(all.z + bx * all.x + by * all.y + all.w);

	    h.x = mad(49.f, i, mad(x, all.x, mad(y, all.y, all.z))) - minh;
	    
//            h.x = mad(49.f, i, mad(sidx, all.x, mad(sidy, all.y, all.w)));
            h.y = h.x + all.x;
            h.z = h.x + all.y;
            h.w = h.y + all.y;
            
            float4 ih = floor(h);
            float4 diff = h - ih;

            v = subh_cache((int)ih.x); v1.x = v.x; v2.x = v.y;
            v = subh_cache((int)ih.y); v1.y = v.x; v2.y = v.y;
            v = subh_cache((int)ih.z); v1.z = v.x; v2.z = v.y;
            v = subh_cache((int)ih.w); v1.w = v.x; v2.w = v.y;

            res += v1  + diff * v2;
	}

        barrier(CLK_LOCAL_MEM_FENCE);
    }

    slice[slice_width * (idy    ) + idx] = res.x;
    slice[slice_width * (idy    ) + idx + 1] = res.y;
    slice[slice_width * (idy + 1) + idx    ] = res.z;
    slice[slice_width * (idy + 1) + idx + 1] = res.w;
}




#undef subh_cache
#define subh_cache(subh) cache[(int)(subh)]

#undef PPT
#define PPT 6
__kernel void hst_backproject_6ppt_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *cache
    ) {
    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = PPT * get_group_id(0) * get_local_size(0);
    const int bidy = PPT * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

/*
    const int sbidx = tidy % 4;
    const int sbidy = tidy / 4;
    
    const int stidx = tidx % 4;
    const int stidy = tidx / 4;

    const int sidx = (sbidx * 4 + stidx);
    const int sidy = (sbidy * 4 + stidy);
*/

    const int sidx = tidx;
    const int sidy = tidy;


    const float sx = sidx;
    const float sy = sidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + off_x;
    const float y = idy + off_y;

    const int slice_width = PPT * get_global_size(0);

    float res[PPT][PPT] = {{0}};
    for (int proj = 0; proj < num_proj; proj++) {
        float4 all = c_all[proj]; all.y = -all.y;
        float minh = floor(all.z + bx * all.x + by * all.y + all.w);

#pragma unroll 5
        for (int i = 0; i < 5; i++) {
	    int pos = i * get_local_size(0) * get_local_size(0) + tidy * get_local_size(0) + tidx;
	    cache[pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + .5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);


        all *= 4;
        float h = mad(x, all.x, mad(y, all.y, all.z)) - 4 * minh;

#pragma unroll 6
        for (int i = 0; i < PPT; i++) {
#pragma unroll 6
            for (int j = 0; j < PPT; j++) {
                res[i][j] += subh_cache(h + (i * get_local_size(0)) * all.y + (j * get_local_size(0)) * all.x);
            }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

#pragma unroll 6
    for (int i = 0; i < PPT; i++) {
#pragma unroll 6
        for (int j = 0; j < PPT; j++) {
            slice[(idy + get_local_size(1) * i) * slice_width +  idx + get_local_size(0) * j] = res[i][j];
        }
    }
}


#undef PPT
#define PPT 8
__kernel void hst_backproject_8ppt_oversample4(const int num_proj, 
    const int num_bins, 
    const float off_x, 
    const float off_y, 
    __read_only image2d_t sinogram,
    __global float *slice,
    __constant float4 *c_all, 
    __local float *cache
    ) {
    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = PPT * get_group_id(0) * get_local_size(0);
    const int bidy = PPT * (get_global_id(1) - get_local_id(1));

    const float bx = bidx + off_x;
    const float by = bidy + off_y;

    const int sidx = tidx;
    const int sidy = tidy;

    const int idx = bidx + sidx;
    const int idy = bidy + sidy;

    const float x = idx + off_x;
    const float y = idy + off_y;

    const int slice_width = PPT * get_global_size(0);

    float res[PPT][PPT] = {{0}};
    for (int proj = 0; proj < num_proj; proj++) {
        float4 all = c_all[proj]; all.y = -all.y;
        float minh = floor(all.z + bx * all.x + by * all.y + all.w);

#pragma unroll 6
        for (int i = 0; i < 6; i++) {
	    int pos = i * get_local_size(0) * get_local_size(0) + tidy * get_local_size(0) + tidx;
	    cache[pos] = read_imagef(sinogram, volumeSampler, (float2)(minh + 0.25f*pos, proj + .5f)).x;
	}

        barrier(CLK_LOCAL_MEM_FENCE);


        all *= 4;
        float h = mad(x, all.x, mad(y, all.y, all.z)) - 4 * minh;

#pragma unroll 8
        for (int i = 0; i < PPT; i++) {
#pragma unroll 8
            for (int j = 0; j < PPT; j++) {
                res[i][j] += subh_cache(h + (i * get_local_size(0)) * all.y + (j * get_local_size(0)) * all.x);
            }
        }

        barrier(CLK_LOCAL_MEM_FENCE);
    }

#pragma unroll 8
    for (int i = 0; i < PPT; i++) {
#pragma unroll 8
        for (int j = 0; j < PPT; j++) {
            slice[(idy + get_local_size(1) * i) * slice_width +  idx + get_local_size(0) * j] = res[i][j];
        }
    }
}

