#define HST_NEWTEX4                             //!< Assymetric: process 4 projections and 64 bins during each step.

//#define HST_SQUARE_PPT				//!< NO. Little effect on performance (slighly reduces register usage, but doesn't affect computations as the loops are unrolled anyway)
//#define HST_FLOAT_LOOPS				//!< NO. Use float loops to optimize register usage (mostly bad)

#ifdef HST_NEWTEX4
# define HST_CACHE_CONST
#endif
//#define HST_NO_OFFSETS
#define HST_NEWTEX_CACHE_BLOCK 128


/*********************************************************************************/
__constant sampler_t volumeSampler = CLK_NORMALIZED_COORDS_FALSE | CLK_ADDRESS_CLAMP | CLK_FILTER_LINEAR;

#if SLICE_BLOCK == 4
# define vfloat float4
# define hst_tex(a,b,c) read_imagef(a, volumeSampler, (float2)(b,c))
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; slice[2 * slice_size + idy*slice_width + idx] = sum.z; slice[3 * slice_size + idy*slice_width + idx] = sum.w; })
#elif SLICE_BLOCK == 2
# define vfloat float2
# define hst_tex(a,b,c) read_imagef(a, volumeSampler, (float2)(b,c)).xy
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum.x; slice[slice_size + idy*slice_width + idx] = sum.y; })
#else
# define vfloat float
# define hst_tex(a,b,c) read_imagef(a, volumeSampler, (float2)(b,c)).x
# define SET_SLICE(slice, idx, idy, sum) ({ slice[idy*slice_width + idx] = sum; })
#endif

/****************************************************************************/

#if defined(HST_NEWTEX4)
# define BIN_DIM 64
# define BIN_STEPS 2
# define BIN_STEP 8
# define PROJ_DIM 4
# if defined(HST_FLOAT_LOOPS)
#  define PROJ_STEP 4.f
# else
#  define PROJ_STEP 4
# endif
# define SHFL_STEP 4
# define PAD_BUF 0
#else /* 16 projections */
# define BIN_DIM 16
# define BIN_STEPS 4
# define BIN_STEP 4
# define PROJ_DIM 16
# if defined(HST_FLOAT_LOOPS)
#  define PROJ_STEP 16.f
# else
#  define PROJ_STEP 16
# endif
# define SHFL_STEP 16
# if  (defined(HST_SHFL_SUM) && (SLICE_BLOCK == 1))
#  define PAD_BUF 4
# else
#  define PAD_BUF 0
# endif
#endif



#undef CACHE_BLOCK
#define CACHE_BLOCK HST_NEWTEX_CACHE_BLOCK

__kernel void newtex(const int num_proj, 
    const int num_bins, 
    const float apos_off_x, 
    const float apos_off_y, 
    __read_only image2d_t texptr,
    __global float *d_SLICE,
    __constant float4 *c_all
) {
    float h;
#if defined(HST_SQUARE_PPT)||!defined(HST_NEWTEX4)
# define SQUARE_PPT
    vfloat res[BIN_STEPS][BIN_STEPS] = {0};
#else
# ifdef HST_NEWTEX4_PPT
#  define YNTEXSIZE (BIN_STEPS * BIN_STEPS * HST_NEWTEX4_PPT)
# else 
#  define YNTEXSIZE (BIN_STEPS * BIN_STEPS)
# endif
# define YNTEXSTEP (BIN_STEP / BIN_STEPS)
    vfloat res[YNTEXSIZE] = {0};
#endif

    __local vfloat buf[BIN_DIM][PROJ_DIM + PAD_BUF]; 
#ifdef HST_NEWTEX4_PPT
    __local vfloat fin[HST_NEWTEX4_PPT * 16][16];
#else
    __local vfloat fin[16][16];
#endif 
    const int tidx = get_local_id(0);
    const int tidy = get_local_id(1);

    const int bidx = get_group_id(0) * get_local_size(0);
    const int bidy = get_group_id(1) * get_local_size(1);

    const int proj_minor = tidx / 4;
    const int xy_minor = tidx % 4;
    const int y_minor = xy_minor / 2;
    const int x_minor = xy_minor % 2;

    const int xy_major = tidy % 4;
    const int y_major = xy_major / 2;
    const int x_major = xy_major % 2;

#if defined(HST_NEWTEX4)
    const int block = tidy / 4;

# ifdef SQUARE_PPT
    const int blocky = block/2;
    const int blockx = block%2;
# else
    const int blocky = 0;
    const int blockx = block;
# endif

    const int ridy = 4 * tidy + tidx / 4;
    const int ridx = tidx % 4;

    const int sidy = blocky * 4 + y_major * 2 + y_minor;
    const int sidx = blockx * 4 + x_major * 2 + x_minor;

    const int proj = proj_minor;
#else
    const int ridy = tidy;
    const int ridx = tidx;

    const int sidy = y_major * 2 + y_minor;
    const int sidx = x_major * 2 + x_minor;

    const int proj_major = tidy / 4;
    const int proj =  4 * proj_major + proj_minor;
#endif

    const float x = bidx + sidx + apos_off_x;
    const float y = bidy + sidy + apos_off_y;

#ifdef HST_NO_OFFSETS
    const float axis = c_all[0].z;
#endif

#ifdef HST_CACHE_CONST
    const int tid = 16 * tidy + tidx;
    __local float2 w_trig[CACHE_BLOCK];
# ifndef HST_NO_OFFSETS
    __local float  w_ofst[CACHE_BLOCK];
# endif

    for (int pblock = 0; pblock < num_proj; pblock += CACHE_BLOCK) {
	const int pblock_end = min(CACHE_BLOCK, num_proj - pblock);
	if (tid < CACHE_BLOCK) {
	    const int g_proj = pblock + tid;
	    w_trig[tid] = (float2){ c_all[g_proj].x, c_all[g_proj].y };
# ifndef HST_NO_OFFSETS
    	    w_ofst[tid] = c_all[g_proj].z;
#endif
	}
	barrier(CLK_LOCAL_MEM_FENCE);

	for (int proje = proj; proje < pblock_end; proje += PROJ_STEP) {
	    const float projf = pblock + proje + 0.5f;
# ifdef HST_NO_OFFSETS
	    const float4 all = (float4){w_trig[proje].x, w_trig[proje].y, axis, 0};
# else
	    const float4 all = (float4){w_trig[proje].x, w_trig[proje].y, w_ofst[proje], 0};
# endif
#else
# if defined(HST_FLOAT_LOOPS)
	for (float projf =  proj + 0.5f; projf < num_proj; projf += PROJ_STEP) {
//	    float fidx = projf + exp2(23.f); const int idx = (*(int*)(&fidx)) - 0x4B000000;
    	    const int proje = (int)projf;
# else
	for (int proje = proj; proje < num_proj; proje += PROJ_STEP) {
	    const float projf = proje + 0.5f;
# endif
	    const float4 all = c_all[proje];		
#endif
	    h = all.z + x * all.x - y * all.y;

#ifdef SQUARE_PPT
# pragma unroll
    	    for (int i = 0; i < BIN_STEPS; i++) {
# pragma unroll
		for (int j = 0; j < BIN_STEPS; j++) {
		    float subh = h + BIN_STEP * j * all.x - BIN_STEP * i * all.y;
		    res[i][j] += hst_tex(texptr, subh, projf);
		}
	    }
#else
# pragma unroll
	    for (int i = 0; i < YNTEXSIZE; i++) {
		float subh = h - YNTEXSTEP * i * all.y;
		res[i] += hst_tex(texptr, subh, projf);
	    }
#endif
    }

#ifdef HST_CACHE_CONST
    }
#endif

#ifdef SQUARE_PPT
# pragma unroll
    for (int i = 0; i < BIN_STEPS; i++) {
# pragma unroll
	for (int j = 0; j < BIN_STEPS; j++) {
            buf[BIN_STEP * sidy + sidx][proj] = res[i][j];
#else
# pragma unroll
    for (int i = 0; i < YNTEXSIZE; i++) {
            buf[16 * sidy + sidx][proj] = res[i]; //?
#endif

            barrier(CLK_LOCAL_MEM_FENCE);

# pragma unroll
            for (int k=(PROJ_DIM/2); k>=1; k/=2) {
                buf[ridy][ridx] += buf[ridy][ridx+k];
            	    // We need it instead of volatile. Could be faster if buf declared volatile
                barrier(CLK_LOCAL_MEM_FENCE);
            }
            vfloat val = buf[ridy][0];

#ifdef SQUARE_PPT
            const int rx = BIN_STEP * j + ridy%BIN_STEP;
            const int ry = BIN_STEP * i + ridy/BIN_STEP;
#else
            const int rx = ridy % 16;
            const int ry = YNTEXSTEP * i + ridy / 16;
#endif

            if (!ridx) {
                fin[ry][rx] = val;
            }

            barrier(CLK_LOCAL_MEM_FENCE);
#ifdef SQUARE_PPT
        }
#endif
    }

    const int idx = bidx + tidx;
    const int idy = bidy + tidy;
    const int slice_width = get_global_size(0);	// Scale by PPT if necessary
    const int slice_size = get_global_size(1) * slice_width; // Scale by PPT if necessary

    SET_SLICE(d_SLICE, idx, idy, fin[tidy][tidx]);
}
